/***************************************************************************
 * Copyright 2012 by
 *  + Christian-Albrechts-University of Kiel
 *    + Department of Computer Science
 *      + Software Engineering Group 
 *  and others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/
package kieker.webgui.common;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.faces.application.FacesMessage;
import javax.faces.context.FacesContext;

import kieker.analysis.AnalysisController;
import kieker.analysis.exception.AnalysisConfigurationException;
import kieker.analysis.model.analysisMetaModel.MIDependency;
import kieker.analysis.model.analysisMetaModel.MIProject;
import kieker.analysis.model.analysisMetaModel.impl.MAnalysisMetaModelFactory;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;

import org.primefaces.model.UploadedFile;

/**
 * This is a singleton class for the access to the file system. It makes sure that the necessary directories for the execution of the application exist. <b>Do
 * not</b> remove directories created from this manager during runtime! Directories are created during first access to the class.<br>
 * Currently nearly all methods are synchronized. A fine grained synchronization should be implemented in the future.
 * 
 * @author Nils Christian Ehmke
 * @version 1.0
 */
public final class FileManager {

	/**
	 * The logger within this class.
	 */
	private static final Log LOG = LogFactory.getLog(FileManager.class);
	/**
	 * The root dir which contains all other directories.
	 */
	private static final String ROOT_DIR = "data";
	/**
	 * The directory which contains the projects.
	 */
	private static final String PROJECT_DIR = FileManager.ROOT_DIR + File.separator + "projects";
	/**
	 * The directory which contains the libraries.
	 */
	private static final String LIB_DIR = FileManager.ROOT_DIR + File.separator + "libraries";
	/**
	 * The extension of the kieker analysis files.
	 */
	private static final String EXTENSION = ".kax";
	/**
	 * The extension of jar-files.
	 */
	private static final String JAR_EXTENSION = ".jar";
	/**
	 * The buffer size in bytes uses to upload dependencies.
	 */
	private static final int BUF_SIZE = 1024;
	/**
	 * The singleton instance of this class.
	 */
	private static final FileManager INSTANCE = new FileManager();
	/**
	 * The factory used to create new objects of the meta model.
	 */
	private final MAnalysisMetaModelFactory factory;

	/**
	 * Creates a new instance of this class.
	 */
	private FileManager() {
		this.checkAndCreateDirectories();
		this.factory = new MAnalysisMetaModelFactory();
	}

	/**
	 * Checks whether all directories are available and creates them if necessary.
	 */
	private void checkAndCreateDirectories() {
		synchronized (this) {
			/*
			 * Make sure that the directories exist and create them if necessary.
			 */
			final File dirProj = new File(FileManager.PROJECT_DIR);
			final File dirLib = new File(FileManager.LIB_DIR);
			boolean couldCreated = true;

			if (!dirProj.exists()) {
				couldCreated &= dirProj.mkdirs();
			}
			if (!dirLib.exists()) {
				couldCreated &= dirLib.mkdirs();
			}

			if (!couldCreated) {
				FileManager.LOG.error("Could not create the necessary directories for the application");
			}
		}
	}

	/**
	 * This method saves a given project using the name of the project as the project-directory. The project-directory must already exist.
	 * 
	 * @param project
	 *            The project to be stored.
	 * @return true iff the project-directory does already exist and the storage was successful.
	 */
	public final boolean saveProject(final MIProject project) {
		synchronized (this) {
			final String projectName = project.getName();

			final File dirProject = new File(FileManager.PROJECT_DIR + File.separator + projectName);

			/*
			 * Make sure that the directory for the project exists.
			 */
			if (dirProject.exists()) {
				/*
				 * Try to save the project.
				 */
				final File fileProject = new File(dirProject, projectName + FileManager.EXTENSION);
				try {
					final AnalysisController controller = new AnalysisController(project, PluginClassLoader.getInstance());
					try {
						controller.saveToFile(fileProject);
						FacesContext.getCurrentInstance().addMessage(null, new FacesMessage(FacesMessage.SEVERITY_INFO, "", "Project saved: " + project.getName()));
						return true;
					} catch (final IOException ex) {
						FileManager.LOG.error("Could not save project '" + projectName + "'.", ex);
					} catch (final AnalysisConfigurationException ex) {
						FileManager.LOG.error("Could not save project '" + projectName + "'.", ex);
					}
				} catch (final AnalysisConfigurationException ex) {
					FileManager.LOG.error("Could not save project '" + projectName + "'.", ex);
				} catch (final NullPointerException ex) {
					FileManager.LOG.error("Could not save project '" + projectName + "'.", ex);
				}
			}
			return false;
		}
	}

	/**
	 * This method saves a given project using the name of the project as the project-directory. The project-directory must not already exist.
	 * 
	 * @param project
	 *            The project to be stored.
	 * @return true iff the project-directory does not already exist and the storage was successful.
	 */
	public final boolean saveNewProject(final MIProject project) {
		synchronized (this) {
			final String projectName = project.getName();
			final File dirProject = new File(FileManager.PROJECT_DIR + File.separator + projectName);

			/*
			 * Make sure that the project does not already exist and create the
			 * project directory.
			 */
			if (dirProject.exists() || !dirProject.mkdir()) {
				return false;
			} else {
				/*
				 * The directory should exist now. Store the project.
				 */
				return this.saveProject(project);
			}
		}
	}

	/**
	 * This method can be used to load all currently saved projects.
	 * 
	 * @return A list containing the loaded projects. If something went wrong, an empty list will be returned, never null.
	 */
	public final List<MIProject> loadAllProjects() {
		final List<MIProject> resultList = new ArrayList<MIProject>();
		synchronized (this) {
			/*
			 * Try to get all directories within the project directory.
			 */
			final File[] directories = new File(FileManager.PROJECT_DIR).listFiles();
			if (directories != null) {
				for (final File directory : directories) {
					if (directory.isDirectory()) {
						/*
						 * If there is a project file within the directory, we know
						 * the name of it.
						 */
						final File projectFile = new File(directory, directory.getName() + FileManager.EXTENSION);
						if (projectFile.exists()) {
							/*
							 * Try to load the project.
							 */
							MIProject project;
							try {
								project = AnalysisController.loadFromFile(projectFile);
								resultList.add(project);
							} catch (final IOException ex) {
								FileManager.LOG.error("Could not load project.", ex);
							}
						}
					}
				}
			}
		}
		return resultList;
	}

	/**
	 * This method uploads a given file as a new dependency. The file is stored within the lib-directory of this application. If a file with the same name does
	 * already exist, it will be replaces.
	 * 
	 * @param file
	 *            The file to be uploaded.
	 * @return The new dependency iff the uploading was successful, null otherwise.
	 */
	public final MIDependency uploadDependency(final UploadedFile file) {
		synchronized (this) {
			final File depFile = new File(FileManager.LIB_DIR, file.getFileName());

			BufferedInputStream in = null;
			BufferedOutputStream out = null;
			try {
				/*
				 * Get the streams.
				 */
				in = new BufferedInputStream(file.getInputstream());
				out = new BufferedOutputStream(new FileOutputStream(depFile));
				final byte[] buf = new byte[FileManager.BUF_SIZE];
				int count;

				/*
				 * Transfer the file.
				 */
				while ((count = in.read(buf)) != -1) {
					out.write(buf, 0, count);
				}

			} catch (final IOException ex) {
				FileManager.LOG.error("Could not transfer file '" + file.getFileName() + "'");
				return null;
			} finally {
				/*
				 * Try to make sure that the streams will be closed.
				 */
				try {
					if (in != null) {
						in.close();
					}
				} catch (final IOException ex) {
					FileManager.LOG.warn("Error while uploading dependency '" + file.getFileName() + "'.");
				}
				try {
					if (out != null) {
						out.close();
					}
				} catch (final IOException ex) {
					FileManager.LOG.warn("Error while uploading dependency '" + file.getFileName() + "'.");
				}
			}
			final MIDependency dependency = this.factory.createDependency();
			dependency.setFilePath(depFile.getName());
			return dependency;
		}
	}

	/**
	 * Delivers the only instance of this class.
	 * 
	 * @return The singleton instance of this class.
	 */
	public static final synchronized FileManager getInstance() {
		return FileManager.INSTANCE;
	}

	/**
	 * This method tries to load all available dependencies from the file system.
	 * 
	 * @return A list containing all available dependencies. If there are no files, an empty list will be returned.
	 */
	public final List<MIDependency> loadAllDependencies() {
		final List<MIDependency> resultList = new ArrayList<MIDependency>();
		synchronized (this) {
			/*
			 * Try to get all files within the library directory.
			 */
			final File[] files = new File(FileManager.LIB_DIR).listFiles();
			if (files != null) {
				for (final File file : files) {
					if (file.isFile() && file.getName().endsWith(FileManager.JAR_EXTENSION)) {
						final MIDependency dependency = this.factory.createDependency();
						dependency.setFilePath(file.getName());
						resultList.add(dependency);
					}
				}
			}
		}
		return resultList;
	}

	/**
	 * Tries to delete a given dependency.
	 * 
	 * @param dependency
	 *            The dependency to be removed.
	 * @return true iff the dependency exists and the removal was successful.
	 */
	public final boolean deleteDependency(final MIDependency dependency) {
		synchronized (this) {
			final File file = new File(FileManager.LIB_DIR, dependency.getFilePath());
			if (file.isFile()) {
				return file.delete();
			}
			return false;
		}
	}

	/**
	 * This method tries to delete a project, by removing the directory of the project and its contents. If something went wrong, the integrity of the folder
	 * structure is not guaranteed.
	 * 
	 * @param project
	 *            The project to be removed.
	 * @return true iff the project has been removed successfully.
	 */
	public final boolean deleteProject(final MIProject project) {
		final String projectName = project.getName();

		final File dirProject = new File(FileManager.PROJECT_DIR + File.separator + projectName);
		synchronized (this) {
			if (dirProject.isDirectory()) {
				final File[] files = dirProject.listFiles();
				for (final File file : files) {
					if (!file.delete()) {
						return false;
					}
				}
				return dirProject.delete();
			}
			return false;
		}
	}

	/**
	 * This method reloads the given project from the file system.
	 * 
	 * @param project
	 *            The project to be reloaded.
	 * @return The reloaded project.
	 */
	public MIProject reloadProject(final MIProject project) {
		final String projectName = project.getName();

		final File projectFile = new File(FileManager.PROJECT_DIR + File.separator + projectName + File.separator + projectName + FileManager.EXTENSION);
		synchronized (this) {
			if (projectFile.isFile()) {
				try {
					return AnalysisController.loadFromFile(projectFile);
				} catch (final IOException ex) {
					FileManager.LOG.error("Could not load project.", ex);
				}
			}
		}
		return null;
	}

	/**
	 * This method can be used to deliver the correct and full path to a given library. This is necessary as only the name of the dependencies is stored within the
	 * model dependency. It is not guaranteed that the given path points to an existing file though.
	 * 
	 * @param dependency
	 *            The dependency which will be used to assemble the path.
	 * @return An absolute path to the given dependency.
	 */
	public String getFullPath(final MIDependency dependency) {
		return new File(FileManager.LIB_DIR, dependency.getFilePath()).getAbsolutePath();
	}
}
