/***************************************************************************
 * Copyright 2012 by
 *  + Christian-Albrechts-University of Kiel
 *    + Department of Computer Science
 *      + Software Engineering Group 
 *  and others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.common;

import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import kieker.analysis.AnalysisController;
import kieker.analysis.model.analysisMetaModel.MIDependency;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;

/**
 * This singleton class is responsible for the dynamic loading of classes. Unlike a normal <code>URLClassLoader</code> it is possible to dynamically add and remove
 * urls from the class loader. This instance should always be used if plugin objects have to be created.
 * 
 * @author Nils Christian Ehmke
 * @version 1.0
 */
public final class PluginClassLoader extends ClassLoader {

	/**
	 * The logger within this class.
	 */
	private static final Log LOG = LogFactory.getLog(PluginClassLoader.class);
	/**
	 * The singleton instance of this class.
	 */
	private static PluginClassLoader instance;

	private final List<URL> list = new ArrayList<URL>();

	private URLClassLoader classLoader = null;

	/**
	 * The default constructor of this class. During the creation all available libraries will be added to the class loader.
	 */
	private PluginClassLoader() {
		this.classLoader = new URLClassLoader(new URL[]{}, AnalysisController.class.getClassLoader());
		/* Make sure that all libs are loaded. */
		final List<MIDependency> libs = FileManager.getInstance().loadAllDependencies();
		for (final MIDependency lib : libs) {
			try {
				this.addURL(new URL("file", "localhost", FileManager.getInstance().getFullPath(lib)));
			} catch (final MalformedURLException ex) {
				PluginClassLoader.LOG.warn("Could not load library.", ex);
			}
		}
	}

	/**
	 * This method can be used to add an url to the class loader.
	 * 
	 * @param url
	 *            The URL of the dependency to be added.
	 */
	public void addURL(final URL url) {
		this.list.add(url);

		this.classLoader = new URLClassLoader(this.list.toArray(new URL[this.list.size()]), AnalysisController.class.getClassLoader());
	}

	/**
	 * This method can be used to remove an url from the class loader.
	 * 
	 * @param url
	 *            The URL of the dependency to be added.
	 */
	public void removeURL(final URL url) {
		this.list.remove(url);

		this.classLoader = new URLClassLoader(this.list.toArray(new URL[this.list.size()]), AnalysisController.class.getClassLoader());
	}

	/**
	 * Delivers the only instance of this class.
	 * 
	 * @return The singleton instance of this class.
	 */
	public static final PluginClassLoader getInstance() {
		/* Create the singleton instance if necessary and use a doPrivileged-block. */
		synchronized (PluginClassLoader.class) {
			if (PluginClassLoader.instance == null) {
				PluginClassLoader.instance = (PluginClassLoader) AccessController.doPrivileged(new PrivilegedAction() {
					@Override
					public Object run() {
						return new PluginClassLoader();
					}
				});
			}
		}
		return PluginClassLoader.instance;
	}

	/**
	 * This method tries to load the class with the given name using the currently loaded dependencies.
	 * 
	 * @param name
	 *            The name of the class to be loaded.
	 * @return The class.
	 * @throws ClassNotFoundException
	 *             If a class with the given name could not be found.
	 */
	@Override
	public Class<?> loadClass(final String name) throws ClassNotFoundException {
		return this.classLoader.loadClass(name);
	}
}
